import plotly.express as px
import pandas as pd


class ImportancePlotter:
    """
    This class is responsible for plotting feature importance results.
    """

    def __init__(self):
        pass

    def plot_metrics(self, df: pd.DataFrame, n_features=None):
        """
        Plots the feature importance results.

        Args:
            n_features (int, optional): Number of top features to plot. Defaults to None.

        Returns:
            None: Displays a Plotly bar chart.
        """
        if n_features:
            df = df[:n_features]
        fig = px.bar(
            df,
            y="Importance",
            labels={"Importance": "Importance Score", "index": "Features tested"},
            title="Feature Importance Plot",
            color="Importance",
            color_continuous_scale="Blues",
        )
        fig.update_traces(marker_line_color="black", marker_line_width=1.5)
        fig.update_traces(hovertemplate="<b>%{y:.4f}</b><extra></extra>")
        fig.show()



import pandas as pd

from typing import List


class DataShifter:
    """
    This class is responsible for shifting data in a DataFrame.

    Args:
        None

    Returns:
        None
    """

    def __init__(self):
        pass

    def shifts_single_column(
        self, df: pd.DataFrame, shifts: List, target_column: str
    ) -> List[pd.DataFrame]:
        """
        Creates shifted versions of the dataset for a single column.

        Args:
            df (pd.DataFrame): Input dataframe.
            target_column (str): Column to apply the shift on.

        Returns:
            List[pd.DataFrame]: List of shifted dataframes.
        """
        data_shifted = []
        for value in shifts:
            df1 = df.copy()
            df1[target_column] = df1[target_column].shift(value)
            df1 = df1.dropna()
            data_shifted.append(df1)
        return data_shifted

    def shift_multiple_columns(
        self, df: pd.DataFrame, columns: List, shifts: List
    ) -> pd.DataFrame:
        """
        Generates shifted datasets for all columns, used for multiprocessing.

        Returns:
            List[Tuple[str, pd.DataFrame]]: List of tuples containing
            the column name and its corresponding shifted dataframe.
        """
        data_shifted = []
        for column in columns:
            for value in shifts:
                df1 = df.copy()
                df1[column] = df1[column].shift(value)
                df1 = df1.dropna()
                data_shifted.append((column, df1))  # Saves column and dataframe
        return data_shifted



from joblib import Parallel, delayed


class FunctionParallizer:
    """
    A class to parallelize function execution using joblib.
    """

    def __init__(self, n_jobs: int = -1):
        self.n_jobs = n_jobs

    def handle(self, func, iterable, *args, **kwargs):
        """
        Parallelizes the execution of a function over an iterable using joblib.
        The only require functionality is to provide a function and an iterable.
        The iterable must contain dataframes.

        Args:
            func (callable): The function to be parallelized. It must accept a dataframe as its first argument.
            iterable (iterable): An iterable containing dataframes to be processed.
            *args: Additional positional arguments to pass to the function.
            **kwargs: Additional keyword arguments to pass to the function.
        """
        return Parallel(n_jobs=self.n_jobs)(
            delayed(func)(df=item, *args, **kwargs) for item in iterable
        )



from featclus.model import FeatureSelection
from featclus.data_processing import DataShifter
from sklearn.pipeline import Pipeline

import pandas as pd


class TestFeatureSelection:
    """
    This class contains unit tests for the FeatureSelection class.
    """

    def test_model_initialization(self):
        """
        Test if the arguments when creating a model instance are defined as expected.
        """
        model = FeatureSelection(
            data=pd.DataFrame(), shifts=[1, 2], n_jobs=2, use_gower=True
        )

        assert isinstance(model.data, pd.DataFrame)
        assert isinstance(model.shifts, list)
        assert isinstance(model.model, Pipeline)
        assert isinstance(model.use_gower, bool)
        assert isinstance(model.n_jobs, int)
        assert model.cache_history == 0
        assert not model.results

    def test_data_shift(self, data_dummy):
        """
        This test is created to test the functionality of shifting a column in pandas.
        """
        model = DataShifter()
        shifted_data = model.shifts_single_column(
            df=data_dummy, target_column="feature_0", shifts=[5]
        )[0]
        assert len(shifted_data) == len(data_dummy) - 5

    def test_shift_df_columns(self, data_dummy):
        """
        This test is created to test the functionality of shifting all columns in a dataframe.
        """
        model = DataShifter()
        shifts = [5, 10]
        result = model.shift_multiple_columns(
            df=data_dummy, shifts=shifts, columns=data_dummy.columns
        )
        expected_number_of_shifts = len(data_dummy.columns) * len(shifts)
        assert len(result) == expected_number_of_shifts

    def test_scoring_normal(self, data_dummy):
        """
        This test is created to test the scoring functionality of the model.
        """
        model = FeatureSelection(data=data_dummy, shifts=[5])
        score = model._get_score(df=data_dummy)
        assert isinstance(score, float)

    def test_scoring_gower(self, data_dummy):
        """
        This test is created to test the scoring functionality of the model using Gower distance.
        """
        model = FeatureSelection(data=data_dummy, shifts=[5], use_gower=True)
        score = model._get_score(df=data_dummy)
        assert isinstance(score, float)

    def test_get_metrics(self, data_dummy):
        """
        This test is created to test the feature permutation over the different shifts.
        """
        model = FeatureSelection(data=data_dummy, shifts=[5], n_jobs=1)
        results = model.get_metrics()
        assert isinstance(results, pd.DataFrame)

    def test_plot_metrics(self, data_dummy):
        """
        This test is created to test the plotting functionality of the model.
        """
        model = FeatureSelection(data=data_dummy, shifts=[5], n_jobs=1)
        model.get_metrics()
        fig = model.plot_results(5)
        assert fig is None

    def test_model_single_core(self, data_dummy):
        """
        This test is created to test the full model functionality using a single core.
        """
        model = FeatureSelection(data=data_dummy, shifts=[5, 10], n_jobs=1)
        results = model.get_metrics()
        assert isinstance(results, pd.DataFrame)

    def test_model_multi_core(self, data_dummy):
        """
        This test is created to test the full model functionality using multiple cores.
        """
        model = FeatureSelection(data=data_dummy, shifts=[5, 10], n_jobs=2)
        results = model.get_metrics()
        assert isinstance(results, pd.DataFrame)



import numpy as np
import pandas as pd

from typing import List, Dict

from sklearn.cluster import DBSCAN
from sklearn.preprocessing import MinMaxScaler
from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline
from sklearn.metrics import silhouette_score

import gower

from .data_processing import DataShifter
from .plots import ImportancePlotter
from .parallelizer import FunctionParallizer


class FeatureSelection:
    """
    This library perforns feature selection for clustering problems.
    The main idea is to shift the data and calculate the silhouette score for each feature.
    The components used are:
        - MinMaxScaler
        - PCA
        - DBSCAN
    The performance of the model is calculated by the silhouette score.

    Args:
        data (pd.DataFrame): The data to be used in the model.
        shifts (List): The shifts to be used in the data.
        n_jobs (int): Number of parallel jobs to run. Default is 1.
        model (Pipeline): The clustering pipeline used for scoring.
        use_gower (bool): Whether to use Gower distance as input for DBSCAN.

    Returns:
        pd.DataFrame: A DataFrame with the importance of each feature sorted.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        shifts: List = [5, 10, 50],
        n_jobs: int = -1,
        model: Pipeline = Pipeline(
            steps=[
                ("scaler", MinMaxScaler()),
                ("pca", PCA(0.8)),
                ("clustering", DBSCAN()),
            ]
        ),
        use_gower: bool = False,
    ):
        self.data = data
        self.shifts = shifts
        self.model = model
        self.use_gower = use_gower
        self.columns = data.columns
        self.n_jobs = n_jobs
        self.cache_history = 0
        self.results = None

        self.data_shifter = DataShifter()
        self.plotter = ImportancePlotter()
        self.parallelizer = FunctionParallizer(n_jobs=n_jobs)

    def _get_permutation_scores(
        self, df: pd.DataFrame, columns: List, shifts: List
    ) -> Dict:
        """
        Calculates the silhouette scores for multiple shifted columns in parallel.

        Args:
            df (pd.DataFrame): Input dataframe.
            columns (List): List of columns to shift.
            shifts (List): List of shifts to apply.

        Returns:
            dict: Dictionary mapping each column to its average silhouette score.
        """
        dataframes = self.data_shifter.shift_multiple_columns(
            df=df, columns=columns, shifts=shifts
        )
        results = self.parallelizer.handle(
            func=self._get_score,
            iterable=[df for _, df in dataframes],
        )
        scores = {col: [] for col in self.columns}
        for (col, _), score in zip(dataframes, results):
            scores[col].append(score)
        return {col: np.mean(scores[col]) for col in scores}

    def _get_score(self, df) -> float:
        """
        Calculates the silhouette score for the given dataframe.

        Args:
            df (pd.DataFrame): Input dataframe.

        Returns:
            float: Silhouette score for the clustering result.
        """
        if self.use_gower:
            labels = self.model.fit_predict(gower.gower_matrix(df.astype("float64")))
        else:
            labels = self.model.fit_predict(df)
        score = silhouette_score(X=df, labels=labels)
        return score

    def get_metrics(self) -> pd.DataFrame:
        """
        Computes feature importance based on the difference between
        shifted and original silhouette scores.

        Returns:
            pd.DataFrame: Sorted DataFrame of feature importance scores.
        """
        scores_shifted = self._get_permutation_scores(
            df=self.data, columns=self.columns, shifts=self.shifts
        )
        original_score = self._get_score(df=self.data)
        final_values = {}
        for key, value in scores_shifted.items():
            final_values[key] = np.abs(original_score - value)
        df = pd.DataFrame(
            final_values.values(), columns=["Importance"], index=final_values.keys()
        ).sort_values("Importance", ascending=False)
        self.cache_history = 1
        self.results = df
        return df

    def plot_results(self, n_features: int = 3):
        """
        Plots the feature importance metrics.

        Args:
            n_features (int): Number of top features to plot. Default is 3.
        """
        if self.cache_history == 0:
            self.get_metrics()
        self.plotter.plot_metrics(df=self.results, n_features=n_features)
